module.exports = {
  presets: [['@babel/preset-env', { targets: { node: 'current' } }]],
  plugins: [
    function () {
      return {
        visitor: {
          MetaProperty (path) {
            path.replaceWithSourceString('process')
          }
        }
      }
    }
  ]
}
