.nl-papaya-container {
    display: none;
    width: 50%;
    position: relative;
}

.nl-papaya-alert {
    margin: .5em !important;
    display: none;
}

.nl-colorbar-container {
    position: absolute;
    width: 22%;
    bottom: 4.8em;
    right: 5em;
}

.nl-colorbar-container .nl-color-bar {
    display: flex;
    flex-direction: row;
}

.nl-colorbar-container .ui.labeled.input {
    margin: 0 -4em;
}

.nl-colorbar-container .ui.labeled.input input {
    width: 4em;
}