body {
  font-family: Avenir, Helvetica, Arial, sans-serif;
  -webkit-font-smoothing: antialiased;
  -moz-osx-font-smoothing: grayscale;
  color: #2c3e50;
}

.main.menu .item img.logo {
  margin-right: 1.5em;
  width: 4em;
}

.main.menu .item .github.icon {
  font-size: 1.5em;
}
